/*
    cc -O3 raw2hdf.c -o raw2hdf -I$HDFINC -L$HDFLIB -lmfhdf -ldf -lz -lm -ljpeg
*/

/**
  * @file raw2hdf.c
  * @brief Command line tool that write binary data in the sds of an HDF file
  *
  * @author Jacques Descloitres, CGTD ICARE
  * @version 1.0.0
  *
  * ----------------------------------------------------------------------------
  *
  * <b>History :</b>
  *
  * ----------------------------------------------------------------------------
  *
  */

#include <stdio.h>
#include "mfhdf.h"

#define MAXNAMELENGTH 200

/**
 * @struct SDS represents a sds
 */
typedef struct {
    char name[MAXNAMELENGTH];
    int32 id, index, num_type, rank, n_attr, Nl, Np, *plane, Nplanes;
    int32 start[MAX_VAR_DIMS], edges[MAX_VAR_DIMS], dim_sizes[MAX_VAR_DIMS];
    void *data, *fillvalue;
} SDS;

/**
 * @brief retrieve an HDF type code using a string representatin of it
 * @param typename the type given in a string ("char8", "float64"...)
 * @return the corresponding HDF type code (DFNT_CHAR8, DFNT_FLOAT64...)
 */
int get_num_type( char* typename ) {
    if ( strcmp(typename,"char8") == 0 )
        return DFNT_CHAR8;
    else if ( strcmp(typename,"uchar8") == 0 )
        return DFNT_UCHAR8;
    else if ( strcmp(typename,"int8") == 0 )
        return DFNT_INT8;
    else if ( strcmp(typename,"uint8") == 0 )
        return DFNT_UINT8;
    else if ( strcmp(typename,"int16") == 0 )
        return DFNT_INT16;
    else if ( strcmp(typename,"uint16") == 0 )
        return DFNT_UINT16;
    else if ( strcmp(typename,"int32") == 0 )
        return DFNT_INT32;
    else if ( strcmp(typename,"uint32") == 0 )
        return DFNT_UINT32;
    else if ( strcmp(typename,"float32") == 0 )
        return DFNT_FLOAT32;
    else if ( strcmp(typename,"float64") == 0 )
        return DFNT_FLOAT64;
    else
        return FAIL;
}

/**
 * @brief command line tool entry point
 * @param argc as usual
 * @param argv[] as usual
 * @return 0 if all works fine, -1 either
 */
int main(int argc, char *argv[]) {
    char filename1[MAXNAMELENGTH], filename2[MAXNAMELENGTH], typename[10];
    int32 sd_id, num_type;
    int j, irow, write_mode, width, height;
    FILE *fd;
    char verbose=FALSE;
    SDS sds;
    void *data;

    filename1[0] = (char)NULL;
    filename2[0] = (char)NULL;
    typename[0] = (char)NULL;
    write_mode = DFACC_CREATE;
    width = height = 0;

    if (argc < 3) {
        fprintf(stderr, "Usage: %s [-v] [-append] -size=<width>x<height> -type=float64|float32|int32|uint32|int16|uint16|int8|uint8|char8|uchar8 -of=<output file> <input file>\n", argv[0]);
        exit(-1);
    }

    for (j=1; j<argc; j++)
        if ( strcmp(argv[j], "-v") == 0 ) {
            fprintf(stderr, "Verbose mode requested.\n");
            verbose = TRUE;
        }

    for (j=1; j<argc; j++) {
        if ( strcmp(argv[j], "-v") == 0 )
            continue;
        else if (strstr(argv[j], "-of=") == argv[j] ) {
            if ( sscanf(argv[j], "-of=%s", filename2) != 1 ) {
                fprintf(stderr, "Cannot parse output file\n");
                exit(-1);
            }
            if (verbose)
                fprintf(stderr, "Output file: %s\n", filename2);
        } else if ( strcmp(argv[j], "-append") == 0 ) {
            write_mode = DFACC_RDWR;
            if (verbose)
                fprintf(stderr, "Append mode\n");
        } else if ( strstr(argv[j], "-type=") == argv[j] ) {
            if ( sscanf(argv[j], "-type=%s", typename) != 1 ) {
                fprintf(stderr, "Cannot parse input data type\n");
                continue;
            }
            if (verbose)
                fprintf(stderr, "Input data type: %s\n", typename);
        } else if ( strstr(argv[j], "-size=") == argv[j] ) {
            if ( sscanf(argv[j], "-size=%dx%d", &width, &height) != 2 ) {
                fprintf(stderr, "Cannot parse input dimensions\n");
                continue;
            }
            if (verbose)
                fprintf(stderr, "Input dimensions: %dx%d\n", width, height);
        } else {
            strcpy(filename1, argv[j]);
            if (verbose)
                fprintf(stderr, "Input file: %s\n", filename1);
        }
    }

    if ( filename1[0] == (char)NULL ) {
        fprintf(stderr, "Input file name not specified\n");
        exit(-1);
    }

    if ( filename2[0] == (char)NULL ) {
        fprintf(stderr, "Output file name not specified\n");
        exit(-1);
    }

    if ( typename[0] == (char)NULL ) {
        fprintf(stderr, "Input data type not specified\n");
        exit(-1);
    }

    num_type=get_num_type(typename);
    if ( num_type == FAIL ) {
        fprintf(stderr, "%s: data type not supported.\n", typename);
        exit(-1);
    }

    if ( width <= 0  ||
            height <= 0 ) {
        fprintf(stderr, "Input dimensions not specified\n");
        exit(-1);
    }

    if ( (fd = fopen(filename1, "r")) == NULL ) {
        fprintf(stderr, "Cannot open file %s.\n", filename1);
        exit(-1);
    }

    if ( write_mode == DFACC_RDWR  &&
            access(filename2, F_OK) == -1 ) {
        fprintf(stderr, "Cannot append file %s. File doesn't exist.\n", filename2);
        exit(-1);
    }

    if ( (sd_id = SDstart(filename2, write_mode)) == -1 ) {
        fprintf(stderr, "Cannot open file %s.\n", filename2);
        exit(-1);
    }

    strcpy(sds.name, filename1);
    sds.rank = 2;
    sds.dim_sizes[0] = height;
    sds.dim_sizes[1] = width;
    sds.num_type = num_type;
    if ((sds.id = SDcreate(sd_id, sds.name, sds.num_type, sds.rank, sds.dim_sizes)) == -1) {
        fprintf(stderr, "  Cannot create SDS \"%s\"\n",  sds.name);
        exit(-1);
    }

    data = (void *) malloc(width * DFKNTsize(num_type));

    for (irow=0; irow<height; irow++) {
        if ( fread(data, DFKNTsize(num_type), width, fd) != width ) {
            fprintf(stderr, "Cannot read line %d from file %s\n", irow, filename1);
            break;
        }
        sds.start[0] = irow;
        sds.start[1] = 0;
        sds.edges[0] = 1;
        sds.edges[1] = width;
        if (SDwritedata(sds.id, sds.start, NULL, sds.edges, data) == -1) {
            fprintf(stderr, "  Cannot write line %d into SDS %s\n", irow, sds.name);
            break;
        }
    }

    free(data);

    fclose(fd);
    SDendaccess(sds.id);
    SDend(sd_id);

    return(0);
}
